/*
 * nibs_mysql.c		SQL Module
 *
 * Version:	$Id: nibs_mysql.c,v 1.14.2.1 2003/07/09 20:46:02 neon Exp $
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright 2000  The FreeRADIUS server project
 * Copyright 2000  Mike Machado <mike@innercite.com>
 * Copyright 2000  Alan DeKok <aland@ox.org>
 */

#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>

#include 	"radiusd.h"

#include "config.h"

#ifdef HAVE_MYSQL_MYSQL_H
#include <mysql/errmsg.h>
#include <mysql/mysql.h>
#else
#ifdef HAVE_MYSQL_H
#include <errmsg.h>
#include <mysql.h>
#endif
#endif

#include	"rlm_nibs.h"

typedef struct rlm_nibs_mysql_sock {
	MYSQL conn;
	MYSQL *sock;
	MYSQL_RES *result;
	SQL_ROW row;
} rlm_nibs_mysql_sock;

/*************************************************************************
 *
 *	Function: als_create_socket
 *
 *	Purpose: Establish connection to the db
 *
 *************************************************************************/
static int als_init_socket(SQLSOCK *sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock;

	if (!sqlsocket->conn) {
		sqlsocket->conn = (rlm_nibs_mysql_sock *)rad_malloc(sizeof(rlm_nibs_mysql_sock));
		if (!sqlsocket->conn) {
			return -1;
		}
	}

	mysql_sock = sqlsocket->conn;
	memset(mysql_sock, 0, sizeof(*mysql_sock));

	radlog(L_INFO, "rlm_nibs_mysql: Starting connect to MySQL server for #%d",
			sqlsocket->id);

	mysql_init(&(mysql_sock->conn));
	if (!(mysql_sock->sock = mysql_real_connect(&(mysql_sock->conn),
						    config->nibs_server,
						    config->nibs_login,
						    config->nibs_password,
						    config->nibs_db,
						    atoi(config->nibs_port),
						    NULL,
						    CLIENT_FOUND_ROWS))) {
		radlog(L_ERR, "rlm_nibs_mysql: Couldn't connect socket to MySQL server %s@%s:%s", config->nibs_login, config->nibs_server, config->nibs_db);
		radlog(L_ERR, "rlm_nibs_mysql: Mysql error '%s'", mysql_error(&mysql_sock->conn));
		mysql_sock->sock = NULL;
		return -1;
	}


	return 0;
}


/*************************************************************************
 *
 *	Function: als_destroy_socket
 *
 *	Purpose: Free socket and any private connection data
 *
 *************************************************************************/
static int als_destroy_socket(SQLSOCK *sqlsocket, NIBS_CONFIG *config)
{
	free(sqlsocket->conn);
	sqlsocket->conn = NULL;

	return 0;
}


/*************************************************************************
 *
 *	Function: als_check_error
 *
 *	Purpose: check the error to see if the server is down
 *
 *************************************************************************/
static int als_check_error(int error)
{
	switch(error) {
	case CR_SERVER_GONE_ERROR:
	case CR_SERVER_LOST:
	case -1:
		radlog(L_DBG, "rlm_nibs_mysql: MYSQL check_error: %d, returning SQL_DOWN", error);
		return SQL_DOWN;
		break;
	case 0:
		return 0;
		break;
	case CR_OUT_OF_MEMORY:
	case CR_COMMANDS_OUT_OF_SYNC:
	case CR_UNKNOWN_ERROR:
	default:
		radlog(L_DBG, "rlm_nibs_mysql: MYSQL check_error: %d received", error);
		return -1;
		break;
	}
}


/*************************************************************************
 *
 *	Function: als_query
 *
 *	Purpose: Issue a query to the database
 *
 *************************************************************************/
static int als_query(SQLSOCK * sqlsocket, NIBS_CONFIG *config, const char *querystr)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	if (config->sqltrace)
		radlog(L_DBG,"rlm_nibs_mysql: query:  %s", querystr);
	if (mysql_sock->sock == NULL) {
		radlog(L_ERR, "rlm_nibs_mysql: Socket not connected");
		return SQL_DOWN;
	}

	mysql_query(mysql_sock->sock, querystr);
	return als_check_error(mysql_errno(mysql_sock->sock));
}


/*************************************************************************
 *
 *	Function: als_store_result
 *
 *	Purpose: database specific store_result function. Returns a result
 *               set for the query.
 *
 *************************************************************************/
static int als_store_result(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	if (mysql_sock->sock == NULL) {
		radlog(L_ERR, "rlm_nibs_mysql: Socket not connected");
		return SQL_DOWN;
	}
	if (!(mysql_sock->result = mysql_store_result(mysql_sock->sock))) {
		radlog(L_ERR, "rlm_nibs_mysql: MYSQL Error: Cannot get result");
		radlog(L_ERR, "rlm_nibs_mysql: MYSQL Error: %s",
		       mysql_error(mysql_sock->sock));
		return als_check_error(mysql_errno(mysql_sock->sock));
	}
	return 0;
}


/*************************************************************************
 *
 *	Function: als_num_fields
 *
 *	Purpose: database specific num_fields function. Returns number
 *               of columns from query
 *
 *************************************************************************/
static int als_num_fields(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	int     num = 0;
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

#if MYSQL_VERSION_ID >= 32224
	if (!(num = mysql_field_count(mysql_sock->sock))) {
#else
	if (!(num = mysql_num_fields(mysql_sock->sock))) {
#endif
		radlog(L_ERR, "rlm_nibs_mysql: MYSQL Error: No Fields");
		radlog(L_ERR, "rlm_nibs_mysql: MYSQL error: %s",
		       mysql_error(mysql_sock->sock));
	}
	return num;
}


/*************************************************************************
 *
 *	Function: als_select_query
 *
 *	Purpose: Issue a select query to the database
 *
 *************************************************************************/
static int als_select_query(SQLSOCK *sqlsocket, NIBS_CONFIG *config,
			    const char *querystr)
{
	int ret;

	ret = als_query(sqlsocket, config, querystr);
	if(ret)
		return ret;
	ret = als_store_result(sqlsocket, config);
	if (ret) {
		return ret;
	}

	/* Why? Per http://www.mysql.com/doc/n/o/node_591.html,
	 * this cannot return an error.  Perhaps just to complain if no
	 * fields are found?
	 */
	als_num_fields(sqlsocket, config);

	return ret;
}


/*************************************************************************
 *
 *	Function: als_num_rows
 *
 *	Purpose: database specific num_rows. Returns number of rows in
 *               query
 *
 *************************************************************************/
static int als_num_rows(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	if (mysql_sock->result)
		return mysql_num_rows(mysql_sock->result);

	return 0;
}


/*************************************************************************
 *
 *	Function: als_fetch_row
 *
 *	Purpose: database specific fetch_row. Returns a SQL_ROW struct
 *               with all the data for the query in 'sqlsocket->row'. Returns
 *		 0 on success, -1 on failure, SQL_DOWN if database is down.
 *
 *************************************************************************/
static int als_fetch_row(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	/*
	 *  Check pointer before de-referencing it.
	 */
	if (!mysql_sock->result) {
		return SQL_DOWN;
	}

	sqlsocket->row = mysql_fetch_row(mysql_sock->result);

	if (sqlsocket->row == NULL) {
		return als_check_error(mysql_errno(mysql_sock->sock));
	}
	return 0;
}


/*************************************************************************
 *
 *	Function: als_free_result
 *
 *	Purpose: database specific free_result. Frees memory allocated
 *               for a result set
 *
 *************************************************************************/
static int als_free_result(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	if (mysql_sock->result) {
		mysql_free_result(mysql_sock->result);
		mysql_sock->result = NULL;
	}

	return 0;
}



/*************************************************************************
 *
 *	Function: als_error
 *
 *	Purpose: database specific error. Returns error associated with
 *               connection
 *
 *************************************************************************/
static const char *als_error(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	if (mysql_sock == NULL || mysql_sock->sock == NULL) {
		return "rlm_nibs_mysql: no connection to db";
	}
	return (const char *)mysql_error(mysql_sock->sock);
}


/*************************************************************************
 *
 *	Function: als_close
 *
 *	Purpose: database specific close. Closes an open database
 *               connection
 *
 *************************************************************************/
static int als_close(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	if (mysql_sock && mysql_sock->sock){
		mysql_close(mysql_sock->sock);
		mysql_sock->sock = NULL;
	}

	return 0;
}


/*************************************************************************
 *
 *	Function: als_finish_query
 *
 *	Purpose: End the query, such as freeing memory
 *
 *************************************************************************/
static int als_finish_query(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	return 0;
}



/*************************************************************************
 *
 *	Function: als_finish_select_query
 *
 *	Purpose: End the select query, such as freeing memory or result
 *
 *************************************************************************/
static int als_finish_select_query(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	als_free_result(sqlsocket, config);

	return 0;
}


/*************************************************************************
 *
 *	Function: als_affected_rows
 *
 *	Purpose: End the select query, such as freeing memory or result
 *
 *************************************************************************/
static int als_affected_rows(SQLSOCK * sqlsocket, NIBS_CONFIG *config)
{
	rlm_nibs_mysql_sock *mysql_sock = sqlsocket->conn;

	return mysql_affected_rows(mysql_sock->sock);
}


/* Exported to rlm_nibs */
rlm_nibs_module_t rlm_nibs_mysql = {
	"rlm_nibs_mysql",
	als_init_socket,
	als_destroy_socket,
	als_query,
	als_select_query,
	als_store_result,
	als_num_fields,
	als_num_rows,
	als_fetch_row,
	als_free_result,
	als_error,
	als_close,
	als_finish_query,
	als_finish_select_query,
	als_affected_rows
};
